#include "raytracer/raytracer.h"
#include "c_cpp_aliases/aliases.h"
#include "vector/vec.h"
#include "window/window.h"
#include <math.h>

#define SPECULAR_EPSILON 0.001f
#define REFLECTIVE_EPSILON 0.1f

INTERNAL intersection_t find_closest_intersection(vec3f_t origin,
                                                  vec3f_t direction, f32 t_min,
                                                  f32 t_max,
                                                  const scene_t *scene);
INTERNAL solutions_t ray_intersects_sphere(vec3f_t origin, vec3f_t direction,
                                           sphere_t sphere);
INTERNAL f32 compute_lighting(vec3f_t position, vec3f_t surface_normal,
                              vec3f_t view_vector, f32 specular_exponent,
                              const scene_t *scene);
INTERNAL f32 light_diffuse(f32 light_intensity, vec3f_t light_direction,
                           vec3f_t surface_normal);
INTERNAL f32 light_specular(f32 light_intensity, vec3f_t light_direction,
                            vec3f_t surface_normal, vec3f_t view_vector,
                            f32 specular_exponent);
INTERNAL vec3f_t reflect_ray(vec3f_t light_direction, vec3f_t surface_normal);
INTERNAL f32 cos_angle_between_vectors(vec3f_t v1, vec3f_t v2);

colour_t trace_ray(vec3f_t origin, vec3f_t direction, f32 t_min, f32 t_max,
                   const scene_t *scene, colour_t default_colour,
                   u32 recursion_depth) {
  intersection_t intersection =
      find_closest_intersection(origin, direction, t_min, t_max, scene);

  if (!intersection.closest_sphere) {
    return default_colour;
  }

  vec3f_t _direction = vec_mul_num(vec3f_t, direction, intersection.closest_t);
  vec3f_t position = vec_add(vec3f_t, origin, _direction);

  vec3f_t surface_normal =
      vec_sub(vec3f_t, position, intersection.closest_sphere->centre);

  f32 normal_magnitude = vec_magnitude(vec3f_t, surface_normal);
  surface_normal = vec_div_num(vec3f_t, surface_normal, normal_magnitude);

  vec3f_t view_vector = vec_mul_num(vec3f_t, direction, -1.0f);

  f32 light = compute_lighting(position, surface_normal, view_vector,
                               intersection.closest_sphere->specular, scene);

  colour_t local_colour =
      colour_mul(intersection.closest_sphere->colour, light);

  f32 reflective = intersection.closest_sphere->reflective;
  if (recursion_depth <= 0 || reflective <= 0) {
    return local_colour;
  }

  vec3f_t R = reflect_ray(view_vector, surface_normal);
  colour_t reflected_colour =
      trace_ray(position, R, REFLECTIVE_EPSILON, INFINITY, scene,
                default_colour, recursion_depth - 1);

  local_colour = colour_mul(local_colour, 1.0f - reflective);
  reflected_colour = colour_mul(reflected_colour, reflective);

  return colour_add_colour(local_colour, reflected_colour);
}

INTERNAL intersection_t find_closest_intersection(vec3f_t origin,
                                                  vec3f_t direction, f32 t_min,
                                                  f32 t_max,
                                                  const scene_t *scene) {
  f32 closest_t = INFINITY;
  sphere_t *closest_sphere = NULL;

  for (u32 i = 0; i < scene->spheres_count; ++i) {
    solutions_t solutions =
        ray_intersects_sphere(origin, direction, scene->spheres[i]);

    if (solutions.t1 >= t_min && solutions.t1 <= t_max &&
        solutions.t1 < closest_t) {
      closest_t = solutions.t1;
      closest_sphere = &(scene->spheres[i]);
    }

    if (solutions.t2 >= t_min && solutions.t2 <= t_max &&
        solutions.t2 < closest_t) {
      closest_t = solutions.t2;
      closest_sphere = &(scene->spheres[i]);
    }
  }

  return (intersection_t){closest_t, closest_sphere};
}

INTERNAL solutions_t ray_intersects_sphere(vec3f_t origin, vec3f_t direction,
                                           sphere_t sphere) {
  f32 r = sphere.radius;
  vec3f_t CO = vec_sub(vec3f_t, origin, sphere.centre);

  f32 a = vec_dot(vec3f_t, direction, direction);
  f32 b = 2.0f * vec_dot(vec3f_t, CO, direction);
  f32 c = vec_dot(vec3f_t, CO, CO) - r * r;

  f32 discriminant = b * b - 4 * a * c;
  if (discriminant < 0) {
    return (solutions_t){INFINITY, INFINITY};
  }

  f32 t1 = (-b + sqrtf(discriminant)) / (2 * a);
  f32 t2 = (-b - sqrtf(discriminant)) / (2 * a);

  return (solutions_t){t1, t2};
}

INTERNAL f32 compute_lighting(vec3f_t position, vec3f_t surface_normal,
                              vec3f_t view_vector, f32 specular_exponent,
                              const scene_t *scene) {
  f32 I = 0.0f;
  light_t light = {0};

  for (u32 i = 0; i < scene->lights_count; ++i) {
    light = scene->lights[i];

    if (light.type == LIGHT_TYPE_AMBIENT) {
      I += light.intensity;
    } else {
      vec3f_t light_direction = {0};
      f32 t_max = SPECULAR_EPSILON;

      switch (light.type) {
      case LIGHT_TYPE_POINT:
        light_direction = vec_sub(vec3f_t, light.position, position);
        t_max = 1;
        break;
      case LIGHT_TYPE_DIRECTIONAL:
        light_direction = light.direction;
        t_max = INFINITY;
        break;
      default:
        break;
      }

      intersection_t shadow = find_closest_intersection(
          position, light_direction, SPECULAR_EPSILON, t_max, scene);
      if (shadow.closest_sphere != NULL) {
        continue;
      }

      I += light_diffuse(light.intensity, light_direction, surface_normal);

      if (specular_exponent != -1.0f) {
        I += light_specular(light.intensity, light_direction, surface_normal,
                            view_vector, specular_exponent);
      }
    }
  }

  return I;
}

INTERNAL f32 light_diffuse(f32 light_intensity, vec3f_t light_direction,
                           vec3f_t surface_normal) {
  return light_intensity *
         cos_angle_between_vectors(light_direction, surface_normal);
}

INTERNAL f32 light_specular(f32 light_intensity, vec3f_t light_direction,
                            vec3f_t surface_normal, vec3f_t view_vector,
                            f32 specular_exponent) {
  vec3f_t R = reflect_ray(light_direction, surface_normal);

  return light_intensity *
         powf(cos_angle_between_vectors(R, view_vector), specular_exponent);
}

INTERNAL vec3f_t reflect_ray(vec3f_t light_direction, vec3f_t surface_normal) {
  vec3f_t _2N = vec_mul_num(vec3f_t, surface_normal, 2.0f);
  f32 dot_product = vec_dot(vec3f_t, light_direction, surface_normal);

  vec3f_t _2N_mul_dot = vec_mul_num(vec3f_t, _2N, dot_product);

  return vec_sub(vec3f_t, _2N_mul_dot, light_direction);
}

INTERNAL f32 cos_angle_between_vectors(vec3f_t v1, vec3f_t v2) {
  f32 dot_product = vec_dot(vec3f_t, v1, v2);

  if (dot_product < 0.0f) {
    return 0.0f;
  }

  f32 divisor = vec_magnitude(vec3f_t, v1) * vec_magnitude(vec3f_t, v2);

  if (divisor == 0.0f) {
    return 0.0f;
  }

  return dot_product / divisor;
}
